# -*- coding: utf-8 -*-
import glob
import law
import luigi
import logging
import os

from itertools import product
from subprocess import PIPE

from law.util import colored

from analysis.tasks.FastProd import FastProd
from analysis.tasks.FnloCppread import FnloCppread

from analysis.framework import Task, SpaceSeparatedListParameter
from analysis.util import interruptable_popen

logger = logging.getLogger(__name__)


class MultiGridClosure(Task, law.LocalWorkflow):
    command = 'fastnnlo_multigridclosure.py'

    force = luigi.BoolParameter(default=False)

    merge_dir = luigi.Parameter(default='grids')
    plots_dir = luigi.Parameter(default='plots')

    default_channels = ["LO", "R", "V", "RRa", "RRb", "RV", "VV"]
    channels = SpaceSeparatedListParameter(
        default=default_channels,
        cls=luigi.Parameter,
        choices=("LO", "R", "V", "RR", "RRa", "RRb", "RV", "VV")
    )
    observables = luigi.ListParameter()
    scalecombs = SpaceSeparatedListParameter()

    def create_branch_map(self):
        """Branch data dicts. Each branch should correspond to exactly one output directory."""
        return {
            i : {'channel': ch, 'index': ch_index, 'observable': obs,
                 'seed': FastProd().starting_seeds[ch_index]}
            for i, ((ch_index, ch), obs) in enumerate(product(
                enumerate(self.channels), self.observables,
            ))
        }

    def workflow_requires(self):
        if self.force:
            return 0
        else:
            return {"fnlocppread": FnloCppread.req(self)}

    def requires(self):
        return {
            'fnlocppread': FnloCppread(branch=self.branch_data['index']),
        }

    def output(self):
        out_dir = os.getenv("ANALYSIS_PATH") or os.getcwd()
        return law.LocalDirectoryTarget(
            '{}/{}/{}/{}/{}/{}'.format(
                out_dir,
                self.plots_dir, self.name, self.__class__.__name__,
                self.branch_data['channel'],
                self.branch_data['observable']
            )
        )

    def run(self):
        try:
            with self.output().temporary_path() as self.temp_output_path:
                os.mkdir(self.temp_output_path)

                tablepath = os.path.join(self.merge_dir, self.name, self.branch_data['channel'])
                logger.debug('Table path is: %s', tablepath)
                # FastProd starting seed for channel
                iseed = self.branch_data['seed']
                # FastProd job number for channel
                njobs = FastProd().fastprod_jobs[self.branch_data['index']]
                fseed = iseed + njobs
                # Find smallest seed for which table exists (and corresponding dat file, see check in FnloCppread)
                ind = iseed
                iseed = -1
                while ind < fseed:
                    tableglob = os.path.join(tablepath,'*.{}.s{}.tab.gz'.format(self.branch_data['observable'],ind))
                    logger.debug("Table glob: %s", tableglob)
                    tables = glob.glob(tableglob)
                    # There should be exactly one matching table
                    if tables:
                        iseed = ind
                        break
                    else:
                        logger.warning('No table found for table glob {} and job seed {}. Trying to find first table in next job!'.format(tableglob,ind))
                        ind = ind+1

                if iseed < 0:
                    message = "No job found for channel {} with a table for observable {}. Multigrid closure plots not possible! Fail!".format(self.branch_data['channel'],self.branch_data['observable'])
                    raise RuntimeError(message)

                # Loop over all matching tables is done in python script; only take the first one
                tabfile = tables[0]
                logger.debug("First table file: %s", tabfile)

                iscl = 0
                sclcoms = self.scalecombs
                if len(sclcoms) > 1:
                    for i, observable in enumerate(self.observables):
                        if observable in tabfile:
                            iscl = i
                            logger.debug(
                                'Table matches observable %s with index %d',
                                observable, iscl
                            )

                sclcom = sclcoms[0]
                if len(sclcoms) > 1:
                    sclcom = sclcoms[iscl]

                outfile = '{}/{}.{}.{}'.format(
                    self.temp_output_path, self.process,
                    self.branch_data['channel'],
                    self.branch_data['observable']
                )

                command_line_args = [
                    self.command, tabfile,
                    '--scale', sclcom,
                    '-f', outfile
                ]

                logger.info("Running command: %s",
                    colored(' '.join(command_line_args), 'yellow'))
                code, out, error = interruptable_popen(command_line_args,
                                                       stdout=PIPE,
                                                       stderr=PIPE)

                logger.debug("Return code: %d", code)
                logger.debug("Standard output:\n%s\n", out)
                logger.debug("Standard error:\n%s\n", error)

                if (code == 0):
                    logger.info("%s status: SUCCESS", self.command)
                else:
                    raise RuntimeError("{} status: FAIL (exit code {})".format(self.command, code))

        # mark job as failed on error, make log entry and reraise
        except Exception as e:
            os.rmdir(self.temp_output_path)
            logger.error('%s raised during job execution: %s', e.__class__.__name__, e)
            raise
