# -*- coding: utf-8 -*-
import law
import luigi
import logging
import os

from itertools import product
from subprocess import PIPE

from law.util import colored

from analysis.tasks.FnloCppreadFinal import FnloCppreadFinal

from analysis.framework import Task, SpaceSeparatedListParameter
from analysis.util import interruptable_popen

logger = logging.getLogger(__name__)


class PdfUncertainty(Task, law.LocalWorkflow):
    command = 'fastnnlo_pdfunc.py'

    merge_dir = luigi.Parameter(default='grids')
    plots_dir = luigi.Parameter(default='plots')

    orders = luigi.ListParameter()
    observables = luigi.ListParameter()

    pdfsets = luigi.Parameter()
    scaleindex = SpaceSeparatedListParameter(cls=luigi.Parameter)

    def create_branch_map(self):
        """Branch data dicts. Each branch should correspond to exactly one output directory."""
        return {
            i : {'order': order}
            for i, order in enumerate(self.orders)
        }

    def run_configs(self):
        """Run configuration dicts. `run` method loops through these to populate the output directory."""
        return [
            {'observable': obs, 'scli': scli}
            for obs in self.observables
            for scli in self.scaleindex
        ]

    def workflow_requires(self):
        return {'fnlocppread': FnloCppreadFinal(self)}

    def output(self):
        out_dir = os.getenv("ANALYSIS_PATH") or os.getcwd()
        return law.LocalDirectoryTarget(
            '{}/{}/{}/{}/{}'.format(
                out_dir,
                self.plots_dir, self.name, self.__class__.__name__,
                self.branch_data['order']
            )
        )

    def run(self):

        try:
            with self.output().temporary_path() as self.temp_output_path:
                os.mkdir(self.temp_output_path)

                for i, run_config in enumerate(self.run_configs()):
                    tabfile = '{}/{}/Combined/Final/{}.{}.{}.tab.gz'.format(
                        self.merge_dir, self.name, self.process,
                        self.branch_data['order'],
                        run_config['observable']
                    )
                    logger.debug("Table file: %s", tabfile)

                    outfile = '{}/{}.{}.{}'.format(
                        self.temp_output_path, self.process,
                        self.branch_data['order'],
                        run_config['observable']
                    )

                    command_line_args = [
                        self.command,
                        '-p', self.pdfsets,
                        '-s', str(run_config['scli']),
                        '-o', self.branch_data['order'],
                        '-f', outfile,
                        tabfile
                    ]

                    logger.info("Running command [%s/%s]: %s", i+1,
                        len(self.run_configs()),
                        colored(' '.join(command_line_args), 'yellow'))
                    code, out, error = interruptable_popen(command_line_args,
                                                           stdout=PIPE,
                                                           stderr=PIPE)

                    logger.debug("Return code: %d", code)
                    logger.debug("Standard output:\n%s\n", out)
                    logger.debug("Standard error:\n%s\n", error)

                    if (code == 0):
                        logger.info("%s status: SUCCESS", self.command)
                    else:
                        raise RuntimeError("{} status: FAIL (exit code {})".format(self.command, code))

        # mark job as failed on error, make log entry and reraise
        except Exception as e:
            os.rmdir(self.temp_output_path)
            logger.error('%s raised during job execution: %s', e.__class__.__name__, e)
            raise
